import logging
from typing import Tuple, Dict

from cirro.cli.interactive.utils import ask_yes_no, ask
from cirro.config import extract_base_url

logger = logging.getLogger()


def gather_auth_config() -> Tuple[str, str, Dict]:
    base_url = ask(
        'text',
        'Enter the URL of the Cirro instance you\'d like to connect to:'
    )
    # Fix user-provided base URL, if necessary
    base_url = extract_base_url(base_url)

    auth_method_config = {
        'enable_cache': ask_yes_no('Would you like to save your login? (do not use this on shared devices)')
    }

    return 'ClientAuth', base_url, auth_method_config



import configparser
import os
import re
from pathlib import Path
from typing import NamedTuple, Dict, Optional

import requests
from requests import RequestException


class Constants:
    home = os.environ.get('CIRRO_HOME', '~/.cirro')
    config_path = Path(home, 'config.ini').expanduser()
    default_base_url = 'cirro.bio'
    default_max_retries = 10


class UserConfig(NamedTuple):
    auth_method: str
    auth_method_config: Dict  # This needs to match the init params of the auth method
    base_url: Optional[str]
    transfer_max_retries: Optional[int]


def extract_base_url(base_url: str):
    # remove http(s):// if it's there
    base_url = re.compile(r'https?://').sub('', base_url).strip()
    # remove anything after the first slash if it's there
    base_url = base_url.split('/')[0]
    return base_url


def save_user_config(user_config: UserConfig):
    original_user_config = load_user_config()
    ini_config = configparser.ConfigParser()
    ini_config['General'] = {
        'auth_method': user_config.auth_method,
        'base_url': user_config.base_url,
        'transfer_max_retries': Constants.default_max_retries
    }
    if original_user_config:
        ini_config['General']['transfer_max_retries'] = str(original_user_config.transfer_max_retries)

    ini_config[user_config.auth_method] = user_config.auth_method_config
    Constants.config_path.parent.mkdir(exist_ok=True)
    with Constants.config_path.open('w') as configfile:
        ini_config.write(configfile)


def load_user_config() -> Optional[UserConfig]:
    if not Constants.config_path.exists():
        return None

    try:
        ini_config = configparser.ConfigParser()
        ini_config.read(str(Constants.config_path.absolute()))
        main_config = ini_config['General']
        auth_method = main_config.get('auth_method')
        base_url = main_config.get('base_url')
        transfer_max_retries = main_config.getint('transfer_max_retries', Constants.default_max_retries)

        if auth_method and ini_config.has_section(auth_method):
            auth_method_config = dict(ini_config[auth_method])
        else:
            auth_method_config = {}

        return UserConfig(
            auth_method=auth_method,
            auth_method_config=auth_method_config,
            base_url=base_url,
            transfer_max_retries=transfer_max_retries
        )
    except Exception:
        raise RuntimeError('Configuration load error, please re-run configuration')


class AppConfig:
    def __init__(self, base_url: str = None):
        self.user_config = load_user_config()
        self.base_url = (base_url or
                         os.environ.get('CIRRO_BASE_URL') or
                         (self.user_config.base_url if self.user_config else None))
        if not self.base_url:
            raise RuntimeError('No base URL provided, please run cirro configure,'
                               ' set the CIRRO_BASE_URL environment variable, or provide the base_url parameter.')
        # remove http(s):// if it's there
        self.base_url = re.compile(r'https?://').sub('', self.base_url).strip()
        self.transfer_max_retries = self.user_config.transfer_max_retries\
            if self.user_config else Constants.default_max_retries
        self._init_config()

    @property
    def checksum_method_display(self):
        return self.checksum_method.replace('NVME', '')

    @property
    def checksum_method(self):
        return 'CRC64NVME'

    def _init_config(self):
        self.rest_endpoint = f'https://{self.base_url}/api'
        self.auth_endpoint = f'https://{self.base_url}/api/auth'

        try:
            info_resp = requests.get(f'{self.rest_endpoint}/info/system')
            info_resp.raise_for_status()

            info = info_resp.json()
            self.client_id = info['auth']['sdkAppId']
            self.user_pool_id = info['auth']['userPoolId']
            self.references_bucket = info['referencesBucket']
            self.resources_bucket = info['resourcesBucket']
            self.region = info['region']
        except RequestException:
            raise RuntimeError(f'Failed connecting to {self.base_url}, please check your configuration')



import logging
from pathlib import Path
from typing import List, Optional, Union, Dict

from cirro_api_client.v1.api.datasets import get_datasets, get_dataset, import_public_dataset, upload_dataset, \
    update_dataset, delete_dataset, get_dataset_manifest
from cirro_api_client.v1.api.sharing import get_shared_datasets
from cirro_api_client.v1.models import ImportDataRequest, UploadDatasetRequest, UpdateDatasetRequest, Dataset, \
    DatasetDetail, CreateResponse, UploadDatasetCreateResponse, FileEntry

from cirro.file_utils import is_hidden_file
from cirro.models.assets import DatasetAssets, Artifact
from cirro.models.dataset import DatasetValidationResponse
from cirro.models.file import FileAccessContext, File, PathLike
from cirro.services.base import get_all_records
from cirro.services.file import FileEnabledService

logger = logging.getLogger()


class DatasetService(FileEnabledService):
    """
    Service for interacting with the Dataset endpoints
    """

    def list(self, project_id: str, max_items: int = 10000) -> List[Dataset]:
        """List datasets

        Retrieves a list of datasets for a given project

        Args:
            project_id (str): ID of the Project
            max_items (int): Maximum number of records to get (default 10,000)
        """
        return get_all_records(
            records_getter=lambda page_args: get_datasets.sync(
                project_id=project_id,
                client=self._api_client,
                next_token=page_args.next_token,
                limit=page_args.limit
            ),
            max_items=max_items
        )

    def list_shared(self, project_id: str, share_id: str, max_items: int = 10000) -> List[Dataset]:
        """
        Retrieves a list of shared datasets for a given project and share

        Args:
            project_id (str): ID of the Project
            share_id (str): ID of the Share
            max_items (int): Maximum number of records to get (default 10,000)

        Example:
        ```python
        from cirro_api_client.v1.models import ShareType
        from cirro.cirro_client import CirroApi

        cirro = CirroApi()

        # List shares that are subscribed to
        subscribed_shares = cirro.shares.list(project_id="project-id", share_type=ShareType.SUBSCRIBER)
        cirro.datasets.list_shared("project-id", subscribed_shares[0].id)
        ```
        """
        return get_all_records(
            records_getter=lambda page_args: get_shared_datasets.sync(
                project_id=project_id,
                share_id=share_id,
                client=self._api_client,
                next_token=page_args.next_token,
                limit=page_args.limit
            ),
            max_items=max_items
        )

    def import_public(self, project_id: str, import_request: ImportDataRequest) -> CreateResponse:
        """
        Download data from public repositories

        Args:
            project_id (str): ID of the Project
            import_request (cirro_api_client.v1.models.ImportDataRequest):

        Returns:
            ID of the created dataset

        ```python
        from cirro_api_client.v1.models import ImportDataRequest, Tag
        from cirro.cirro_client import CirroApi

        cirro = CirroApi()
        request = ImportDataRequest(
            name="Imported dataset",
            description="Description of the dataset",
            public_ids=["SRR123456", "SRR123457"],
            tags=[Tag(value="tag1")]
        )
        cirro.datasets.import_public("project-id", request)
        ```
        """
        return import_public_dataset.sync(project_id=project_id, client=self._api_client, body=import_request)

    def create(self, project_id: str, upload_request: UploadDatasetRequest) -> UploadDatasetCreateResponse:
        """
        Registers a dataset in Cirro, which can subsequently have files uploaded to it

        Args:
            project_id (str): ID of the Project
            upload_request (cirro_api_client.v1.models.UploadDatasetRequest):

        Returns:
            ID of the created dataset and the path to upload files

        ```python
        from cirro_api_client.v1.models import UploadDatasetRequest, Tag
        from cirro.cirro_client import CirroApi

        cirro = CirroApi()
        request = UploadDatasetRequest(
            name="Name of new dataset",
            process_id="paired_dnaseq",
            expected_files=["read_1.fastq.gz", "read_2.fastq.gz"],
            description="Description of the dataset",
            tags=[Tag(value="tag1"), Tag(value="tag2")]
        )
        cirro.datasets.create("project-id", request)
        ```
        """
        return upload_dataset.sync(project_id=project_id, client=self._api_client, body=upload_request)

    def get(self, project_id: str, dataset_id: str) -> Optional[DatasetDetail]:
        """
        Gets detailed information about a dataset

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset

        Returns:
            The dataset, if found
        """
        return get_dataset.sync(project_id=project_id, dataset_id=dataset_id, client=self._api_client)

    def update(self, project_id: str, dataset_id: str, request: UpdateDatasetRequest) -> DatasetDetail:
        """
        Update info on a dataset (name, description, and/or tags)

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            request (cirro_api_client.v1.models.UpdateDatasetRequest):

        Returns:
            The updated dataset

        ```python
        from cirro_api_client.v1.models import UpdateDatasetRequest
        from cirro.cirro_client import CirroApi

        cirro = CirroApi()
        request = UpdateDatasetRequest(
            name="Name of new dataset",
            process_id="paired_dnaseq",
            description="Description of the dataset"
        )
        cirro.datasets.update("project-id", "dataset-id", request)
        ```
        """
        return update_dataset.sync(project_id=project_id, dataset_id=dataset_id, body=request, client=self._api_client)

    def delete(self, project_id: str, dataset_id: str) -> None:
        """
        Delete a dataset

        After a dataset has been deleted, the files associated with that
        dataset are saved according to the project's retention time.

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
        """
        delete_dataset.sync_detailed(project_id=project_id, dataset_id=dataset_id, client=self._api_client)

    def get_assets_listing(self, project_id: str, dataset_id: str, file_limit: int = 100000) -> DatasetAssets:
        """
        Gets a listing of files, charts, and other assets available for the dataset

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            file_limit (int): Maximum number of files to get (default 100,000)
        """
        dataset = self.get(project_id, dataset_id)
        if file_limit < 1:
            raise ValueError("file_limit must be greater than 0")
        all_files = []
        file_offset = 0
        domain = None
        artifacts = None

        while len(all_files) < file_limit:
            manifest = get_dataset_manifest.sync(
                project_id=project_id,
                dataset_id=dataset_id,
                file_offset=file_offset,
                client=self._api_client
            )
            all_files.extend(manifest.files)
            file_offset += len(manifest.files)

            if not artifacts:
                artifacts = manifest.artifacts

            domain = manifest.domain
            if len(all_files) >= manifest.total_files or len(manifest.files) == 0:
                break

        files = [
            File.from_file_entry(
                f,
                project_id=project_id,
                dataset=dataset,
                domain=domain
            )
            for f in all_files
        ]
        artifacts = [
            Artifact(
                artifact_type=a.type,
                file=File.from_file_entry(
                    FileEntry(a.path),
                    project_id=project_id,
                    dataset=dataset,
                    domain=domain
                )
            )
            for a in artifacts
        ]
        return DatasetAssets(files=files, artifacts=artifacts)

    def upload_files(self,
                     project_id: str,
                     dataset_id: str,
                     directory: PathLike,
                     files: List[PathLike] = None,
                     file_path_map: Dict[PathLike, str] = None) -> None:
        """
        Uploads files to a given dataset from the specified directory.

        All files must be relative to the specified directory.
        If files need to be flattened, or you are sourcing files from multiple directories,
        please include `file_path_map` or call this method multiple times.

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            directory (str|Path): Path to directory
            files (typing.List[str|Path]): List of paths to files within the directory,
                must be the same type as directory.
            file_path_map (typing.Dict[str|Path, str|Path]): Optional mapping of file paths to upload
             from source path to destination path, used to "re-write" paths within the dataset.
        ```python
        from cirro.cirro_client import CirroApi
        from cirro.file_utils import generate_flattened_file_map

        cirro = CirroApi()

        directory = "~/Downloads"
        # Re-write file paths
        file_map = {
            "data1/file1.fastq.gz": "file1.fastq.gz",
            "data2/file2.fastq.gz": "file2.fastq.gz",
            "file3.fastq.gz": "new_file3.txt"
        }

        # Or you could automate the flattening
        files = ["data1/file1.fastq.gz"]
        file_map = generate_flattened_file_map(files)

        cirro.datasets.upload_files(
            project_id="project-id",
            dataset_id="dataset-id",
            directory=directory,
            files=list(file_map.keys()),
            file_path_map=file_map
        )
        ```
        """
        if file_path_map is None:
            file_path_map = {}

        dataset = self.get(project_id, dataset_id)

        access_context = FileAccessContext.upload_dataset(
            project_id=project_id,
            dataset_id=dataset_id,
            base_url=dataset.s3
        )

        self._file_service.upload_files(
            access_context=access_context,
            directory=directory,
            files=files,
            file_path_map=file_path_map
        )

    def validate_folder(
        self,
        project_id: str,
        dataset_id: str,
        local_folder: PathLike
    ) -> DatasetValidationResponse:
        """
        Validates that the contents of a dataset match that of a local folder.
        """
        ds_files = self.get_assets_listing(project_id, dataset_id).files

        local_folder = Path(local_folder)
        if not local_folder.is_dir():
            raise ValueError(f"{local_folder} is not a valid local folder")

        # Keep track of files from the dataset which match by checksum, don't match, or are missing
        ds_files_matching = []
        ds_files_not_matching = []
        ds_files_missing = []
        ds_validate_failed = []
        for ds_file in ds_files:
            ds_file_path = ds_file.normalized_path
            # Get the corresponding local file
            local_file = local_folder / ds_file_path
            if not local_file.exists():
                ds_files_missing.append(ds_file_path)
            else:
                try:
                    if self._file_service.is_valid_file(ds_file, local_file):
                        ds_files_matching.append(ds_file_path)
                    else:
                        ds_files_not_matching.append(ds_file_path)
                except RuntimeWarning as e:
                    logger.warning(f"File validation failed: {e}")
                    ds_validate_failed.append(ds_file_path)

        # Find local files that are not in the dataset
        local_file_paths = [
            file.relative_to(local_folder).as_posix()
            for file in local_folder.rglob("*")
            if not file.is_dir() and not is_hidden_file(file)
        ]
        dataset_file_paths = [file.normalized_path for file in ds_files]
        local_only_files = [
            file
            for file in local_file_paths
            if file not in dataset_file_paths
        ]

        return DatasetValidationResponse(
            files_matching=ds_files_matching,
            files_not_matching=ds_files_not_matching,
            files_missing=ds_files_missing,
            local_only_files=local_only_files,
            validate_errors=ds_validate_failed,
        )

    def download_files(
        self,
        project_id: str,
        dataset_id: str,
        download_location: str,
        files: Union[List[File], List[str]] = None
    ) -> None:
        """
        Downloads files from a dataset

        The `files` argument is used to optionally specify a subset of files
        to be downloaded. By default, all files are downloaded.

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            download_location (str): Local destination for downloaded files
            files (typing.List[str]): Optional list of files to download
        """
        if files is None:
            files = self.get_assets_listing(project_id, dataset_id).files

        if len(files) == 0:
            return

        first_file = files[0]
        if isinstance(first_file, File):
            files = [file.relative_path for file in files]
            access_context = first_file.access_context
        else:
            dataset = self.get(project_id, dataset_id)
            if dataset.share:
                access_context = FileAccessContext.download_shared_dataset(project_id=project_id,
                                                                           dataset_id=dataset_id,
                                                                           base_url=dataset.s3)
            else:
                access_context = FileAccessContext.download(project_id=project_id,
                                                            base_url=dataset.s3)

        self._file_service.download_files(access_context, download_location, files)

    def update_samplesheet(
        self,
        project_id: str,
        dataset_id: str,
        samplesheet: str
    ):
        """
        Updates a samplesheet on a dataset

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            samplesheet (str): Samplesheet contents to update (should be a CSV string)
        """
        dataset = self.get(project_id, dataset_id)
        access_context = FileAccessContext.upload_sample_sheet(project_id=project_id,
                                                               dataset_id=dataset_id,
                                                               base_url=dataset.s3)

        samplesheet_key = f'{access_context.prefix}/samplesheet.csv'
        self._file_service.create_file(
            access_context=access_context,
            key=samplesheet_key,
            contents=samplesheet,
            content_type='text/csv'
        )



from pathlib import Path
from typing import List, Optional

from cirro_api_client.v1.api.processes import get_processes, get_process, get_process_parameters, \
    validate_file_requirements, archive_custom_process, create_custom_process, sync_custom_process, \
    update_custom_process
from cirro_api_client.v1.models import ValidateFileRequirementsRequest, Executor, Process, ProcessDetail, \
    CustomPipelineSettings, CustomProcessInput, CreateResponse

from cirro.models.form_specification import ParameterSpecification
from cirro.services.base import BaseService


class ProcessService(BaseService):
    """
    Service for interacting with the Process endpoints
    """
    def list(self, process_type: Executor = None) -> List[Process]:
        """
        Retrieves a list of available processes

        Args:
            process_type (`cirro_api_client.v1.models.Executor`): Optional process type (INGEST, CROMWELL, or NEXTFLOW)
        """
        processes = get_processes.sync(client=self._api_client)
        return [p for p in processes if not process_type or process_type == p.executor]

    def get(self, process_id: str) -> ProcessDetail:
        """
        Retrieves detailed information on a process

        Args:
            process_id (str): Process ID
        """
        return get_process.sync(process_id=process_id, client=self._api_client)

    def archive(self, process_id: str):
        """
        Removes a custom process from the list of available processes.

        Error will be raised if the requested process does not exist. No value
        is returned, and no error raised if process exists and request is satisfied.

        Args:
            process_id (str): Process ID
        """
        archive_custom_process.sync_detailed(process_id=process_id, client=self._api_client)

    def find_by_name(self, name: str) -> Optional[ProcessDetail]:
        """
        Get a process by its display name

        Args:
            name (str): Process name
        """
        matched_process = next((p for p in self.list() if p.name == name), None)
        if not matched_process:
            return None

        return self.get(matched_process.id)

    def get_parameter_spec(self, process_id: str) -> ParameterSpecification:
        """
        Gets a specification used to describe the parameters used in the process

        Args:
            process_id (str): Process ID
        """
        form_spec = get_process_parameters.sync(process_id=process_id, client=self._api_client)
        return ParameterSpecification(form_spec)

    def check_dataset_files(self, files: List[str], process_id: str, directory: str):
        """
        Checks if the file mapping rules for a process are met by the list of files

        Error will be raised if the file mapping rules for the process are not met.
        No value is returned and no error is raised if the rules are satisfied.

        Args:
            process_id (str): ID for the process containing the file mapping rules
            directory: path to directory containing files
            files (List[str]): File names to check
        """
        # Parse sample sheet file if present
        sample_sheet = None
        sample_sheet_file = Path(directory, 'samplesheet.csv')
        if sample_sheet_file.exists():
            sample_sheet = sample_sheet_file.read_text()

        request = ValidateFileRequirementsRequest(
            file_names=files,
            sample_sheet=sample_sheet
        )
        requirements = validate_file_requirements.sync(process_id=process_id, body=request, client=self._api_client)

        # These will be sample sheet errors or no files errors
        if error_msg := requirements.error_msg:
            raise ValueError(error_msg)

        errors = [
            f'{entry.description}. {entry.error_msg}. We accept any of the following naming conventions: \n\t- ' +
            '\n\t- '.join([
                e.example_name
                for e in entry.allowed_patterns
            ])
            for entry in requirements.allowed_data_types
            if entry.error_msg is not None
        ]

        files_provided = ', '.join(files)

        if len(errors) != 0:
            raise ValueError(f"The files you have provided are: {files_provided} \n\n"
                             "They do not meet the dataset requirements. "
                             "The required file types are: \n" +
                             "\n".join(errors))

    def create_custom_process(self, process: CustomProcessInput) -> CreateResponse:
        """
        Creates a custom process (pipeline or data type) in the system.

        Whether it is a pipeline or data type is determined by the executor field.
        For pipelines, you must specify the pipeline code and configuration repository.

        This process will be available to all users in the system if `is_tenant_wide` is set to True.
        If `is_tenant_wide` is set to False, the process will only be available the projects
         specified in `linked_projects_ids`. Making it available tenant wide requires tenant admin privileges.

        If the repository is private, you must complete the authorization flow on the UI.

        See https://docs.cirro.bio/pipelines/importing-custom-pipelines/ for more info.

        Args:
            process (Process): Process to create

        Example:
        ```python
        from cirro_api_client.v1.models import CustomProcessInput, Executor, \
         PipelineCode, FileMappingRule, FileNamePattern, RepositoryType, CustomPipelineSettings
        from cirro.cirro_client import CirroApi

        cirro = CirroApi()

        # New pipeline
        new_pipeline = CustomProcessInput(
            id="my_pipeline",
            name="My Pipeline",
            description="This is a test pipeline",
            executor=Executor.CROMWELL,
            category="DNA Sequencing",
            child_process_ids=[],
            parent_process_ids=["rnaseq"],
            documentation_url="https://example.com/docs",
            pipeline_code=PipelineCode(
                repository_path="CirroBio/test-pipeline",
                version="v1.0.0",
                entry_point="main.nf",
                repository_type=RepositoryType.GITHUB_PUBLIC
            ),
            linked_project_ids=[],
            is_tenant_wide=True,
            allow_multiple_sources=True,
            uses_sample_sheet=True,
            # This can be the same or different from the pipeline_code
            custom_settings=CustomPipelineSettings(
                repository="CirroBio/test-pipeline",
                branch="v1.0.0",
                repository_type=RepositoryType.GITHUB_PUBLIC,
            ),
            file_mapping_rules=[
                FileMappingRule(
                    description="Filtered Feature Matrix",
                    file_name_patterns=[
                        FileNamePattern(
                            example_name="filtered_feature_bc_matrix.h5",
                            description="Matrix",
                            sample_matching_pattern="(?P<sampleName>[\\S ]*)\\.jpg"
                        )
                    ]
                )
            ]
        )

        cirro.processes.create_custom_process(new_pipeline)

        # New data type
        new_data_type = CustomProcessInput(
            id="images_jpg",
            name="JPG Images",
            description="Used for generic JPG images",
            executor=Executor.INGEST,
            child_process_ids=[],
            parent_process_ids=[],
            documentation_url="https://example.com/docs",
            linked_project_ids=["project_id_1", "project_id_2"],
            file_mapping_rules=[
                FileMappingRule(
                    description="Images",
                    min_=1,
                    file_name_patterns=[
                        FileNamePattern(
                            example_name="image.jpg",
                            description="JPG Image",
                            sample_matching_pattern="(?P<sampleName>[\\S ]*)/outs/image\\.jpg"
                        )
                    ]
                )
            ]
        )
        cirro.processes.create_custom_process(new_data_type)
        ```
        """
        return create_custom_process.sync(client=self._api_client, body=process)

    def update_custom_process(self, process_id: str, process: CustomProcessInput):
        """
        Updates the custom process (pipeline or data type) in the system.

        Please run `sync_custom_process` after updating to ensure the pipeline configuration is up to date.

        Args:
            process_id (str): ID of the process to update
            process (CustomProcessInput): Process to update
        """
        update_custom_process.sync_detailed(client=self._api_client,
                                            process_id=process_id,
                                            body=process)

    def sync_custom_process(self, process_id: str) -> CustomPipelineSettings:
        """
        Syncs a custom pipeline in the system.

        This updates the pipeline configurations in Cirro (form configuration,
         input mapping, preprocess etc.) to match what is in the configured repository.
        See https://docs.cirro.bio/pipelines/configuring-pipeline/ for more details.

        Args:
            process_id (str): ID of the process to sync
        """
        return sync_custom_process.sync(client=self._api_client, process_id=process_id)



from .billing import BillingService
from .compute_environment import ComputeEnvironmentService
from .dataset import DatasetService
from .execution import ExecutionService
from .file import FileService
from .metadata import MetadataService
from .metrics import MetricsService
from .process import ProcessService
from .projects import ProjectService
from .references import ReferenceService
from .share import ShareService
from .user import UserService
from .workspace import WorkspaceService


__all__ = [
    'BillingService',
    'DatasetService',
    'ExecutionService',
    'ComputeEnvironmentService',
    'FileService',
    'MetadataService',
    'MetricsService',
    'ProcessService',
    'ProjectService',
    'ReferenceService',
    'ShareService',
    'UserService',
    'WorkspaceService',
]



from io import StringIO

from cirro_api_client.v1.api.datasets import get_sample_sheets, ingest_samples
from cirro_api_client.v1.api.processes import validate_file_name_patterns
from cirro_api_client.v1.models import SampleSheets, ValidateFileNamePatternsRequest, FileNameMatch

from cirro.cirro_client import CirroApi
from cirro.helpers import PreprocessDataset


class Matches(list[FileNameMatch]):
    def print(self):
        """
        Prints the file name validation matches in a readable format.
        """
        print(f'Matches: {len(self)}')
        print()
        for match in self:
            print(f'{match.file_name}')
            print(f'Sample name: {match.sample_name}')
            print(f'Matched regex: {match.regex_pattern_match}')
            print()


class DeveloperHelper:
    """
    Helper class for developer-related tasks,
    such as adding samplesheet preprocessing for a pipeline
    or testing file name validation and sample autopopulation.
    """

    def __init__(self, client: CirroApi):
        self.client = client

    def generate_preprocess_for_input_datasets(self,
                                               project_id: str,
                                               input_dataset_ids: list[str],
                                               params=None) -> PreprocessDataset:
        """
        Generates a PreprocessDataset object for the given datasets

        With optional parameters to pass into the preprocess script.
        `metadata` is not available in this context, so it is mocked.
        """
        samplesheets = self._generate_samplesheets_for_datasets(project_id, input_dataset_ids)
        return PreprocessDataset(
            samplesheet=samplesheets.samples,
            files=samplesheets.files,
            params=params or {},
            # Mock metadata
            metadata={
                'dataset': {},
                'project': {},
                'inputs': [],
                'process': {}
            }
        )

    def test_file_name_validation_for_dataset(self,
                                              project_id: str,
                                              dataset_id: str,
                                              file_name_patterns: list[str]) -> Matches:
        """
        Tests the file name validation for a given dataset against specified regex patterns.

        Used when configuring Cirro's sample autopopulation feature.
        More info: https://docs.cirro.bio/features/samples/#using-auto-population
        """
        dataset_files = self.client.datasets.get_assets_listing(project_id=project_id, dataset_id=dataset_id).files
        file_names = [file.relative_path for file in dataset_files]
        return self.test_file_name_validation(file_names, file_name_patterns)

    def test_file_name_validation(self,
                                  file_names: list[str],
                                  file_name_patterns: list[str]) -> Matches:
        """
        Tests the file name validation for a list of file names against specified regex patterns.
        """
        request_body = ValidateFileNamePatternsRequest(
            file_names=file_names,
            file_name_patterns=file_name_patterns
        )

        matches = validate_file_name_patterns.sync(
            process_id="test",
            body=request_body,
            client=self.client.api_client
        )
        return Matches(matches)

    def generate_samplesheets_for_dataset(self, project_id: str, dataset_id: str) -> SampleSheets:
        """
        Generates Cirro samplesheets for a given dataset
        """
        return get_sample_sheets.sync(
            project_id=project_id,
            dataset_id=dataset_id,
            client=self.client.api_client
        )

    def rerun_sample_ingest_for_dataset(self, project_id: str, dataset_id: str):
        """
        Reruns the sample ingest process for a given dataset.
        You'll want to do this if you have updated the file name patterns in your pipeline (or data type)
        """
        ingest_samples.sync_detailed(
            project_id=project_id,
            dataset_id=dataset_id,
            client=self.client.api_client
        )

    def _generate_samplesheets_for_datasets(self, project_id: str, dataset_ids: list[str]) -> SampleSheets:
        """
        Generates Cirro samplesheets for multiple datasets in a project.
        """
        # Concatenate samplesheets using pandas
        import pandas
        samplesheets_dfs = []
        files_dfs = []
        for dataset_id in dataset_ids:
            samplesheet = self.generate_samplesheets_for_dataset(project_id, dataset_id)
            samplesheets_dfs.append(pandas.read_csv(StringIO(samplesheet.samples)))
            files_dfs.append(pandas.read_csv(StringIO(samplesheet.files)))

        samplesheets_df = pandas.concat(samplesheets_dfs, ignore_index=True)
        files_df = pandas.concat(files_dfs, ignore_index=True)
        return SampleSheets(
            samples=samplesheets_df.to_csv(index=False),
            files=files_df.to_csv(index=False)
        )



from cirro.cli.controller import run_ingest, run_download, run_configure, run_list_datasets
from cirro.cli.controller import run_create_pipeline_config, run_validate_folder

__all__ = [
    'run_ingest',
    'run_download',
    'run_configure',
    'run_list_datasets',
    'run_create_pipeline_config',
    'run_validate_folder'
]



from cirro.models.process.process import PipelineDefinition, ConfigAppStatus, get_input_params, CONFIG_APP_URL

__all__ = [
    'PipelineDefinition',
    'ConfigAppStatus',
    'get_input_params',
    'CONFIG_APP_URL'
]



import datetime
from pathlib import Path
from typing import Union, List, Optional

from cirro_api_client.v1.api.processes import validate_file_requirements
from cirro_api_client.v1.models import Dataset, DatasetDetail, RunAnalysisRequest, ProcessDetail, Status, \
    RunAnalysisRequestParams, Tag, ArtifactType, NamedItem, ValidateFileRequirementsRequest

from cirro.cirro_client import CirroApi
from cirro.models.assets import DatasetAssets
from cirro.models.file import PathLike
from cirro.sdk.asset import DataPortalAssets, DataPortalAsset
from cirro.sdk.exceptions import DataPortalAssetNotFound
from cirro.sdk.exceptions import DataPortalInputError
from cirro.sdk.file import DataPortalFile, DataPortalFiles
from cirro.sdk.helpers import parse_process_name_or_id
from cirro.sdk.process import DataPortalProcess


class DataPortalDataset(DataPortalAsset):
    """
    Datasets in the Data Portal are collections of files which have
    either been uploaded directly, or which have been output by
    an analysis pipeline or notebook.
    """

    def __init__(self, dataset: Union[Dataset, DatasetDetail], client: CirroApi):
        """
        Instantiate a dataset object

        Should be invoked from a top-level constructor, for example:

        ```python
        from cirro import DataPortal()
        portal = DataPortal()
        dataset = portal.get_dataset(
            project="id-or-name-of-project",
            dataset="id-or-name-of-dataset"
        )
        ```

        """
        assert dataset.project_id is not None, "Must provide dataset with project_id attribute"
        self._data = dataset
        self._assets: Optional[DatasetAssets] = None
        self._client = client

    @property
    def id(self) -> str:
        """Unique identifier for the dataset"""
        return self._data.id

    @property
    def name(self) -> str:
        """Editable name for the dataset"""
        return self._data.name

    @property
    def description(self) -> str:
        """Longer name for the dataset"""
        return self._data.description

    @property
    def process_id(self) -> str:
        """Unique ID of process used to create the dataset"""
        return self._data.process_id

    @property
    def process(self) -> ProcessDetail:
        """
        Object representing the process used to create the dataset
        """
        return self._client.processes.get(self.process_id)

    @property
    def project_id(self) -> str:
        """ID of the project containing the dataset"""
        return self._data.project_id

    @property
    def status(self) -> Status:
        """
        Status of the dataset
        """
        return self._data.status

    @property
    def source_dataset_ids(self) -> List[str]:
        """IDs of the datasets used as sources for this dataset (if any)"""
        return self._data.source_dataset_ids

    @property
    def source_datasets(self) -> List['DataPortalDataset']:
        """
        Objects representing the datasets used as sources for this dataset (if any)
        """
        return [
            DataPortalDataset(
                dataset=self._client.datasets.get(project_id=self.project_id, dataset_id=dataset_id),
                client=self._client
            )
            for dataset_id in self.source_dataset_ids
        ]

    @property
    def params(self) -> dict:
        """
        Parameters used to generate the dataset
        """
        return self._get_detail().params.to_dict()

    @property
    def info(self) -> dict:
        """
        Extra information about the dataset
        """
        return self._get_detail().info.to_dict()

    @property
    def tags(self) -> List[Tag]:
        """
        Tags applied to the dataset
        """
        return self._data.tags

    @property
    def share(self) -> Optional[NamedItem]:
        """
        Share associated with the dataset, if any.
        """
        return self._get_detail().share

    @property
    def created_by(self) -> str:
        """User who created the dataset"""
        return self._data.created_by

    @property
    def created_at(self) -> datetime.datetime:
        """Timestamp of dataset creation"""
        return self._data.created_at

    def _get_detail(self):
        if not isinstance(self._data, DatasetDetail):
            self._data = self._client.datasets.get(project_id=self.project_id, dataset_id=self.id)
        return self._data

    def _get_assets(self):
        if not self._assets:
            self._assets = self._client.datasets.get_assets_listing(
                project_id=self.project_id,
                dataset_id=self.id
            )
        return self._assets

    def __str__(self):
        return '\n'.join([
            f"{i.title()}: {self.__getattribute__(i)}"
            for i in ['name', 'id', 'description', 'status']
        ])

    def get_file(self, relative_path: str) -> DataPortalFile:
        """
        Get a file from the dataset using its relative path.

        Args:
            relative_path (str): Relative path of file within the dataset

        Returns:
            `from cirro.sdk.file import DataPortalFile`
        """

        # Get the list of files in this dataset
        files = self.list_files()

        # Try getting the file using the relative path provided by the user
        try:
            return files.get_by_id(relative_path)
        except DataPortalAssetNotFound:
            # Try getting the file with the 'data/' prefix prepended
            try:
                return files.get_by_id("data/" + relative_path)
            except DataPortalAssetNotFound:
                # If not found, raise the exception using the string provided
                # by the user, not the data/ prepended version (which may be
                # confusing to the user)
                msg = '\n'.join([f"No file found with path '{relative_path}'."])
                raise DataPortalAssetNotFound(msg)

    def list_files(self) -> DataPortalFiles:
        """
        Return the list of files which make up the dataset.
        """
        files = self._get_assets().files
        return DataPortalFiles(
            [
                DataPortalFile(file=file, client=self._client)
                for file in files
            ]
        )

    def get_artifact(self, artifact_type: ArtifactType) -> DataPortalFile:
        """
        Get the artifact of a particular type from the dataset
        """
        artifacts = self._get_assets().artifacts
        artifact = next((a for a in artifacts if a.artifact_type == artifact_type), None)
        if artifact is None:
            raise DataPortalAssetNotFound(f"No artifact found with type '{artifact_type}'")
        return DataPortalFile(file=artifact.file, client=self._client)

    def list_artifacts(self) -> List[DataPortalFile]:
        """
        Return the list of artifacts associated with the dataset

        An artifact may be something generated as part of the analysis or other process.
        See `cirro_api_client.v1.models.ArtifactType` for the list of possible artifact types.

        """
        artifacts = self._get_assets().artifacts
        return DataPortalFiles(
            [
                DataPortalFile(file=artifact.file, client=self._client)
                for artifact in artifacts
            ]
        )

    def download_files(self, download_location: str = None) -> None:
        """
        Download all the files from the dataset to a local directory.

        Args:
            download_location (str): Path to local directory
        """

        # Alias for internal method
        self.list_files().download(download_location)

    def run_analysis(
            self,
            name: str = None,
            description: str = "",
            process: Union[DataPortalProcess, str] = None,
            params=None,
            notifications_emails: List[str] = None,
            compute_environment: str = None,
            resume_dataset_id: str = None
    ) -> str:
        """
        Runs an analysis on a dataset, returns the ID of the newly created dataset.

        The process can be provided as either a DataPortalProcess object,
        or a string which corresponds to the name or ID of the process.

        Args:
            name (str): Name of newly created dataset
            description (str): Description of newly created dataset
            process (DataPortalProcess or str): Process to run
            params (dict): Analysis parameters
            notifications_emails (List[str]): Notification email address(es)
            compute_environment (str): Name or ID of compute environment to use,
             if blank it will run in AWS
            resume_dataset_id (str): ID of dataset to resume from, used for caching task execution.
             It will attempt to re-use the previous output to minimize duplicate work

        Returns:
            dataset_id (str): ID of newly created dataset
        """
        if name is None:
            raise DataPortalInputError("Must specify 'name' for run_analysis")
        if process is None:
            raise DataPortalInputError("Must specify 'process' for run_analysis")
        if notifications_emails is None:
            notifications_emails = []
        if params is None:
            params = {}

        # If the process is a string, try to parse it as a process name or ID
        process = parse_process_name_or_id(process, self._client)

        if compute_environment:
            compute_environments = self._client.compute_environments.list_environments_for_project(
                project_id=self.project_id
            )
            compute_environment = next(
                (env for env in compute_environments
                 if env.name == compute_environment or env.id == compute_environment),
                None
            )
            if compute_environment is None:
                raise DataPortalInputError(f"Compute environment '{compute_environment}' not found")

        resp = self._client.execution.run_analysis(
            project_id=self.project_id,
            request=RunAnalysisRequest(
                name=name,
                description=description,
                process_id=process.id,
                source_dataset_ids=[self.id],
                params=RunAnalysisRequestParams.from_dict(params),
                notification_emails=notifications_emails,
                resume_dataset_id=resume_dataset_id,
                compute_environment_id=compute_environment.id if compute_environment else None
            )
        )
        return resp.id

    def update_samplesheet(self,
                           contents: str = None,
                           file_path: PathLike = None):
        """
        Updates the samplesheet metadata of a dataset.
        Provide either the contents (as a string) or a file path.
        Both must be in the format of a CSV.

        Args:
            contents (str): Samplesheet contents to update (should be a CSV string)
            file_path (PathLike): Path of file to update (should be a CSV file)

        Example:
        ```python
        dataset.update_samplesheet(
            file_path=Path('~/samplesheet.csv')
        )
        ```
        """

        if contents is None and file_path is None:
            raise DataPortalInputError("Must specify either 'contents' or 'file_path' when updating samplesheet")

        samplesheet_contents = contents
        if file_path is not None:
            samplesheet_contents = Path(file_path).expanduser().read_text()

        # Validate samplesheet
        file_names = [f.file_name for f in self.list_files()]
        request = ValidateFileRequirementsRequest(
            file_names=file_names,
            sample_sheet=samplesheet_contents,
        )
        requirements = validate_file_requirements.sync(process_id=self.process_id,
                                                       body=request,
                                                       client=self._client.api_client)
        if error_msg := requirements.error_msg:
            raise DataPortalInputError(error_msg)

        # Update the samplesheet if everything looks ok
        self._client.datasets.update_samplesheet(
            project_id=self.project_id,
            dataset_id=self.id,
            samplesheet=samplesheet_contents
        )


class DataPortalDatasets(DataPortalAssets[DataPortalDataset]):
    """Collection of multiple DataPortalDataset objects."""
    asset_name = "dataset"



from typing import Dict, Any, List

import jsonschema
from cirro_api_client.v1.models import FormSchema


def _get_fields_in_schema(schema: Dict, parent_path='') -> List['Parameter']:
    # This breaks for more advanced json schema usages such as the if/then
    fields = []

    for [field_key, field_spec] in schema.items():
        field_path = field_key if parent_path == '' else f'{parent_path}.{field_key}'
        field = Parameter(field_key, field_spec, field_path)
        fields.append(field)

        if 'properties' in field_spec:
            fields.extend(_get_fields_in_schema(field_spec['properties'], field_path))

    return fields


class ParameterSpecification:
    """
    Used to describe parameters used in a process (uses JSONSchema)
    """
    def __init__(self, form_schema: FormSchema):
        self._form_spec_raw: Dict = form_schema.form.additional_properties
        self._form_spec_ui: Dict = form_schema.ui.additional_properties
        self.form_spec = _get_fields_in_schema(self._form_spec_raw.get('properties') or {})

    @property
    def form_spec_json(self) -> dict:
        """
        Returns the JSON schema of the form specification
        """
        return self._form_spec_raw

    def validate_params(self, params: Dict):
        """
        Validates that the given parameters conforms to the specification
        """
        try:
            jsonschema.validate(instance=params, schema=self._form_spec_raw)
        except jsonschema.ValidationError as e:
            raise RuntimeError(f'Parameter at {e.json_path} error: {e.message}') from e

    def print(self):
        """
        Prints out a user-friendly view of the parameters
        """
        print("Parameters:")
        for field in self.form_spec:
            tab_prefix = '\t' * (field.path.count('.') + 1)
            print(tab_prefix + str(field))


class Parameter:
    def __init__(self, key: str, spec: Dict[str, Any], path: str):
        self.spec = spec
        self.key = key
        self.path = path

    @property
    def is_group(self):
        return self.spec.get('type') == 'object'

    def __str__(self):
        display_value = ''
        additional_data = ''

        if self.spec.get('title'):
            display_value += f'{self.spec.get("title")}'
            additional_data += f'key={self.key}, '
        else:
            display_value += self.key

        for prop in ['default', 'type', 'enum', 'description']:
            if prop_value := self.spec.get(prop):
                additional_data += f'{prop}={prop_value}, '

        if additional_data and not self.is_group:
            display_value += f' ({additional_data.rstrip(", ")})'

        if self.is_group:
            display_value += ' (Group)'
        return display_value



from datetime import datetime

import jwt
from cirro_api_client import TokenAuth
from cirro_api_client.cirro_auth import AuthMethod

from cirro.auth.base import AuthInfo


class AccessTokenAuth(AuthInfo):
    """
    Authenticates to Cirro with a static access token

    :param token: Access token
    """

    def __init__(self, token: str):
        self._token = token
        self._username = None
        self._access_token_expiry = None
        self._update_token_metadata()

    def get_current_user(self) -> str:
        return self._username

    def get_auth_method(self) -> AuthMethod:
        return TokenAuth(token=self._token)

    def _update_token_metadata(self):
        decoded_access_token = jwt.decode(self._token,
                                          options={"verify_signature": False})
        self._access_token_expiry = datetime.fromtimestamp(decoded_access_token['exp'])
        self._username = decoded_access_token['username']



import sys
from fnmatch import fnmatch
from pathlib import Path
from typing import List

from cirro_api_client.v1.models import Process, Project
from prompt_toolkit.shortcuts import CompleteStyle

from cirro.cli.interactive.common_args import ask_project
from cirro.cli.interactive.utils import ask, prompt_wrapper, InputError, DirectoryValidator
from cirro.cli.models import UploadArguments
from cirro.file_utils import get_files_in_directory, get_files_stats


def ask_data_directory(input_value: str) -> str:
    directory_prompt = {
        'type': 'path',
        'name': 'data_directory',
        'message': 'Enter the full path of the data directory',
        'validate': DirectoryValidator,
        'default': input_value or '',
        'complete_style': CompleteStyle.READLINE_LIKE,
        'only_directories': True
    }

    answers = prompt_wrapper(directory_prompt)
    return str(Path(answers['data_directory']).expanduser())


def ask_name(input_value: str) -> str:
    name_prompt = {
        'type': 'input',
        'name': 'name',
        'message': 'What is the name of this dataset?',
        'validate': lambda val: len(val.strip()) > 0 or 'Please enter a name',
        'default': input_value or ''
    }

    answers = prompt_wrapper(name_prompt)
    return answers['name']


def ask_description(input_value: str) -> str:
    description_prompt = {
        'type': 'input',
        'name': 'description',
        'message': 'Enter a description of the dataset (optional)',
        'default': input_value or ''
    }

    answers = prompt_wrapper(description_prompt)
    return answers['description']


def confirm_data_files(data_directory: str, files: List[str]):
    stats = get_files_stats([
        Path(data_directory) / file
        for file in files
    ])

    if not ask(
        "confirm",
        f'Please confirm that you wish to upload {stats.number_of_files} files ({stats.size_friendly})'
    ):
        sys.exit(1)


def ask_process(processes: List[Process], input_value: str) -> str:
    process_names = [process.name for process in processes]
    process_names.sort()
    return ask(
        'select',
        'What type of files?',
        default=input_value if input_value in process_names else None,
        choices=process_names,
        use_shortcuts=len(process_names) < 30
    )


def ask_files_in_directory(data_directory: str, include_hidden: bool) -> List[str]:
    # Get the list of all files in the directory
    # (relative to the data_directory)
    files = get_files_in_directory(
        data_directory,
        include_hidden=include_hidden
    )

    choices = [
        "Upload all files",
        "Select files from a list",
        "Select files with a naming pattern (glob)"
    ]

    choice = ask(
        'select',
        'Which files would you like to upload from this dataset?',
        choices=choices
    )

    if choice == choices[0]:
        return files
    elif choice == choices[1]:
        return ask_dataset_files_list(files)
    else:
        return ask_dataset_files_glob(files)


def ask_dataset_files_list(files: List[str]) -> List[str]:
    return prompt_wrapper({
        'type': 'checkbox',
        'name': 'files',
        'message': 'Select the files to upload',
        'choices': files
    })['files']


def ask_dataset_files_glob(files: List[str]) -> List[str]:

    selected_files = ask_dataset_files_glob_single(files)
    confirmed = ask(
        "confirm",
        f'Number of files selected: {len(selected_files):} / {len(files):,}'
    )
    while not confirmed:
        selected_files = ask_dataset_files_glob_single(files)
        confirmed = ask(
            "confirm",
            f'Number of files selected: {len(selected_files):} / {len(files):,}'
        )

    if len(selected_files) == 0:
        raise InputError("No files selected")

    return selected_files


def ask_dataset_files_glob_single(files: List[str]) -> List[str]:

    print("All Files:")
    for file in files:
        print(f" - {file}")

    answers = prompt_wrapper({
        'type': 'text',
        'name': 'glob',
        'message': 'Select files by naming pattern (using the * wildcard)',
        'default': '*'
    })

    selected_files = [
        file
        for file in files
        if fnmatch(file, answers['glob'])
    ]

    print("Selected Files:")
    for file in selected_files:
        print(f" - {file}")

    return selected_files


def gather_upload_arguments(input_params: UploadArguments, projects: List[Project], processes: List[Process]):
    input_params['project'] = ask_project(projects, input_params.get('project'))

    input_params['data_directory'] = ask_data_directory(input_params.get('data_directory'))
    files = ask_files_in_directory(input_params['data_directory'], input_params['include_hidden'])

    confirm_data_files(input_params['data_directory'], files)

    input_params['data_type'] = ask_process(processes, input_params.get('data_type'))

    data_directory_name = Path(input_params['data_directory']).name
    default_name = input_params.get('name') or data_directory_name
    input_params['name'] = ask_name(default_name)
    input_params['description'] = ask_description(input_params.get('description'))
    return input_params, files



from pathlib import Path
from typing import Optional

from cirro_api_client.v1.models import ReferenceType, ReferenceTypeValidationItem

from cirro.models.file import PathLike
from cirro.models.reference import ReferenceValidation


def format_expected_file(validation: ReferenceTypeValidationItem):
    save_as = validation.additional_properties.get('saveAs')
    if glob := validation.additional_properties.get('glob'):
        return f'{glob} (e.g., {save_as})'
    file_type = validation.additional_properties.get('fileType')
    return f'.{file_type} (e.g., {save_as})'


def get_matching_validation(file_name: str,
                            validations: list[ReferenceTypeValidationItem]) -> Optional[ReferenceValidation]:
    for validation in validations:
        validation_dict: ReferenceValidation = validation.to_dict()
        glob_pattern = validation_dict.get('glob')
        if glob_pattern and Path(file_name).match(validation_dict.get('glob')):
            return validation_dict
        if file_name.endswith(validation_dict['fileType']):
            return validation_dict
    return None


def _get_reference_upload_file_name(file: PathLike, reference_type: ReferenceType) -> str:
    if not isinstance(file, Path):
        file = Path(file)
    file_name = file.name

    matching_validation = get_matching_validation(file_name=file_name, validations=reference_type.validation)
    return matching_validation.get('saveAs') \
        if matching_validation and matching_validation.get('saveAs') \
        else file_name


def generate_reference_file_path_map(files: list[PathLike],
                                     name: str,
                                     ref_type: ReferenceType):
    """
    Generate a mapping of reference files to their upload paths based on the reference type.

    Files are uploaded to an expected directory structure based on the reference type.
    For example, <reference_id>/<reference_name>/<normalized_file_name>
    """
    upload_prefix = Path(ref_type.directory, name)

    file_path_map = {
        reference_file: (upload_prefix / _get_reference_upload_file_name(reference_file, ref_type)).as_posix()
        for reference_file in files
    }
    return file_path_map



from typing import List

from attrs import define
from cirro_api_client.v1.models import ArtifactType

from cirro.models.file import File


@define
class Artifact:
    """
    An artifact is a file that is a secondary file associated with the dataset.
    This could be a workflow execution report, a log file, or any other file generated as a result of an analysis.
    """
    artifact_type: ArtifactType
    file: File


@define
class DatasetAssets:
    """
    Container for assets associated with a dataset (files, artifacts, etc.)
    """
    files: List[File]
    artifacts: List[Artifact]



from cirro_api_client.v1.api.workspaces import get_workspaces, get_workspace, get_workspace_environments, \
    create_workspace, delete_workspace, start_workspace, stop_workspace
from cirro_api_client.v1.models import Workspace, WorkspaceInput, CreateResponse, WorkspaceEnvironment

from cirro.services.base import BaseService


class WorkspaceService(BaseService):
    """
    Service for interacting with the Workspace endpoints
    """
    def list_environments(self) -> list[WorkspaceEnvironment]:
        """
        List available workspace environments
        """
        return get_workspace_environments.sync(client=self._api_client)

    def list(self, project_id: str) -> list[Workspace]:
        """
        Retrieves a list of workspaces that the user has access to

        Args:
            project_id (str): ID of the Project
        """
        return get_workspaces.sync(project_id, client=self._api_client)

    def get(self, project_id: str, workspace_id: str) -> Workspace:
        """
        Get details of a workspace

        Args:
            project_id (str): ID of the Project
            workspace_id (str): ID of the Workspace
        """
        return get_workspace.sync(project_id=project_id, workspace_id=workspace_id, client=self._api_client)

    def create(self, project_id: str, workspace: WorkspaceInput) -> CreateResponse:
        """
        Create a new workspace in the given project

        Args:
            project_id (str): ID of the Project
            workspace (WorkspaceInput): Workspace object to create
        """
        return create_workspace.sync(project_id=project_id, client=self._api_client, body=workspace)

    def delete(self, project_id: str, workspace_id: str) -> None:
        """
        Delete a workspace in the given project

        Args:
            project_id (str): ID of the Project
            workspace_id (str): ID of the Workspace
        """
        delete_workspace.sync_detailed(project_id=project_id, workspace_id=workspace_id, client=self._api_client)

    def start(self, project_id: str, workspace_id: str) -> None:
        """
        Start a workspace environment

        Args:
            project_id (str): ID of the Project
            workspace_id (str): ID of the Workspace
        """
        start_workspace.sync_detailed(project_id=project_id, workspace_id=workspace_id, client=self._api_client)

    def stop(self, project_id: str, workspace_id: str) -> None:
        """
        Stop a workspace environment

        Args:
            project_id (str): ID of the Project
            workspace_id (str): ID of the Workspace
        """
        stop_workspace.sync_detailed(project_id=project_id, workspace_id=workspace_id, client=self._api_client)



from urllib.parse import urlparse


class S3Path:
    def __init__(self, url):
        self._parsed = urlparse(url, allow_fragments=False)

    @property
    def bucket(self):
        return self._parsed.netloc

    @property
    def key(self):
        return self._parsed.path.lstrip('/')

    @property
    def valid(self):
        return self._parsed.scheme == 's3'

    @property
    def base(self):
        return f's3://{self.bucket}'

    def __str__(self):
        return self._parsed.geturl()



import json
import logging
import sys
import threading
import time
from datetime import datetime, timedelta
from io import StringIO
from pathlib import Path
from typing import Optional
from typing import TYPE_CHECKING

import boto3
import jwt
import requests
from botocore.exceptions import ClientError
from cirro_api_client.cirro_auth import AuthMethod, RefreshableTokenAuth

if TYPE_CHECKING:
    from msal_extensions.persistence import BasePersistence

from cirro.auth.base import AuthInfo
from cirro.auth.oauth_models import DeviceTokenResponse, OAuthTokenResponse
from cirro.config import Constants

logger = logging.getLogger()


def _build_token_persistence(location: str, fallback_to_plaintext=False):
    try:
        if sys.platform.startswith('win'):
            from msal_extensions import FilePersistenceWithDataProtection
            return FilePersistenceWithDataProtection(location)
        if sys.platform.startswith('darwin'):
            from msal_extensions import KeychainPersistence
            return KeychainPersistence(location, service_name='cirro-client')
        if sys.platform.startswith('linux'):
            from msal_extensions import LibsecretPersistence
            return LibsecretPersistence(location)
        raise RuntimeError(f"Unsupported platform: {sys.platform}")
    except Exception:
        if not fallback_to_plaintext:
            raise
        from msal_extensions import FilePersistence
        logger.debug("Encryption unavailable. Opting in to plain text.")
        return FilePersistence(location)


def _initialize_auth_flow(client_id: str, auth_endpoint: str) -> DeviceTokenResponse:
    params = {'client_id': client_id}
    resp = requests.post(f'{auth_endpoint}/device-code', params=params)
    resp.raise_for_status()
    flow: DeviceTokenResponse = resp.json()
    return flow


def _await_completion(client_id: str, auth_endpoint: str, flow: DeviceTokenResponse):
    device_expiry = datetime.fromisoformat(flow['expiry'])

    params = {
        'client_id': client_id,
        'device_code': flow['device_code'],
        'grant_type': 'urn:ietf:params:oauth:grant-type:device_code'
    }

    auth_status = 'authorization_pending'
    while auth_status == 'authorization_pending':
        time.sleep(flow['interval'])
        if device_expiry < datetime.now().astimezone():
            raise RuntimeError('Authentication timed out')

        resp = requests.post(f'{auth_endpoint}/token', params=params)
        token_result: OAuthTokenResponse = resp.json()
        auth_status = token_result.get('message')
        logger.debug(auth_status)

        if 'access_token' in token_result:
            return token_result

    raise RuntimeError(f'error authenticating {auth_status}')


def _authenticate(client_id: str, auth_endpoint: str, auth_io: Optional[StringIO] = None):
    flow = _initialize_auth_flow(client_id=client_id, auth_endpoint=auth_endpoint)
    if auth_io is None:
        print(flow['message'])
    else:
        auth_io.write(flow['message'])
    return _await_completion(client_id=client_id, auth_endpoint=auth_endpoint, flow=flow)


class DeviceCodeAuth(AuthInfo):
    """
    Authenticates to Cirro by asking
    the user to enter a verification code on the portal website

    :param client_id: The client ID for the OAuth application
    :param region: The AWS region where the Cognito user pool is located
    :param auth_endpoint: The endpoint for the OAuth authorization server
    :param enable_cache: Optionally enable cache to avoid re-authentication
    :param auth_io: Optionally provide a StringIO object for the authentication link
    :param await_completion:
        If True, block until the user completes the authorization.
            If auth_io is provided, the authorization message will be written to that buffer.
            If auth_io is not provided, the authorization message will be printed.
        If False, the object will be instantiated without fully completing the authorization.
            The authorization message can be accessed using the .auth_message property.
            Then, the await_completion() method must be run to complete the process.

    Implements the OAuth device code flow
    This is the preferred way to authenticate
    """
    def __init__(
        self,
        client_id: str,
        region: str,
        auth_endpoint: str,
        enable_cache=False,
        auth_io: Optional[StringIO] = None,
        await_completion=True
    ):
        self.client_id = client_id
        self.auth_endpoint = auth_endpoint
        self.region = region
        self._token_info: Optional[OAuthTokenResponse] = None
        self._persistence: Optional[BasePersistence] = None
        self._flow: Optional[DeviceTokenResponse] = None
        self._token_path = Path(Constants.home, f'{client_id}.token.dat').expanduser()

        if enable_cache:
            self._persistence = _build_token_persistence(str(self._token_path), fallback_to_plaintext=True)
            self._token_info = self._load_token_info()

        # Check saved token for change in endpoint
        if self._token_info and self._token_info.get('client_id') != client_id:
            logger.debug('Different client ID found, clearing saved token info')
            self._clear_token_info()

        # Check saved token for refresh token expiry
        if self._token_info and self._token_info.get('refresh_expires_in'):
            refresh_expiry_threshold = datetime.fromtimestamp(self._token_info.get('refresh_expires_in'))\
                                       - timedelta(hours=12)
            if refresh_expiry_threshold < datetime.now():
                logger.debug('Refresh token expiry is too soon, re-authenticating')
                self._clear_token_info()

        if not self._token_info:
            if await_completion:
                self._token_info = _authenticate(client_id=client_id, auth_endpoint=auth_endpoint, auth_io=auth_io)
            else:
                self._flow = _initialize_auth_flow(client_id=client_id, auth_endpoint=auth_endpoint)

        if self._token_info:
            self._save_token_info()
            self._update_token_metadata()
            self._get_token_lock = threading.Lock()

    @property
    def auth_message(self):
        """
        If the DeviceCodeAuth was instantiated with await_completion=False,
        then the authorization message will be populated by this property.
        """
        if self._flow is None:
            raise ValueError("The DeviceTokenResponse is not available")
        else:
            return self._flow["message"]

    @property
    def auth_message_markdown(self):
        """
        Markdown syntax for the authorization message, so that links are rendered appropriately.
        """
        return " ".join([
            (
                f"[{part}]({part})"
                if part.startswith("http")
                else part
            )
            for part in self.auth_message.split(" ")
        ])

    def await_completion(self):
        """Block until the user completes the authorization process."""
        self._token_info = _await_completion(
            client_id=self.client_id,
            auth_endpoint=self.auth_endpoint,
            flow=self._flow
        )
        self._save_token_info()
        self._update_token_metadata()
        self._get_token_lock = threading.Lock()

    def get_auth_method(self) -> AuthMethod:
        return RefreshableTokenAuth(token_getter=lambda: self._get_token()['access_token'])

    def get_current_user(self) -> str:
        return self._username

    def _get_token(self):
        with self._get_token_lock:
            # Refresh access token using refresh token
            if datetime.now() > self.access_token_expiry:
                self._refresh_access_token()

        return self._token_info

    def _refresh_access_token(self):
        try:
            cognito = boto3.client('cognito-idp', region_name=self.region)
            resp = cognito.initiate_auth(
                ClientId=self.client_id,
                AuthFlow='REFRESH_TOKEN_AUTH',
                AuthParameters={
                    'REFRESH_TOKEN': self._token_info['refresh_token']
                }
            )
            logger.debug('Successfully refreshed token')
        except ClientError as err:
            logger.warning(err)
            self._clear_token_info()
            raise RuntimeError('Failed to refresh token, please reauthenticate')

        auth_result = resp['AuthenticationResult']
        self._token_info['access_token'] = auth_result['AccessToken']
        self._token_info['id_token'] = auth_result['IdToken']
        self._save_token_info()
        self._update_token_metadata()

    def _update_token_metadata(self):
        decoded_access_token = jwt.decode(self._token_info['access_token'],
                                          options={"verify_signature": False})
        self.access_token_expiry = datetime.fromtimestamp(decoded_access_token['exp'])
        self._username = decoded_access_token['username']

    def _load_token_info(self) -> Optional[OAuthTokenResponse]:
        if not self._persistence or not self._token_path.exists():
            return None

        token_info = json.loads(self._persistence.load())
        if 'access_token' not in token_info:
            return None

        return token_info

    def _save_token_info(self):
        if not self._persistence:
            return

        self._persistence.save(json.dumps(self._token_info))

    def _clear_token_info(self):
        if not self._persistence:
            return

        Path(self._persistence.get_location()).unlink(missing_ok=True)
        self._token_info = None



from typing import List

from cirro_api_client.v1.models import Project, Dataset

from cirro.cli.interactive.utils import ask, prompt_wrapper, InputError
from cirro.models.dataset import DatasetWithShare
from cirro.utils import format_date


def _format_share(dataset: Dataset | DatasetWithShare) -> str:
    if isinstance(dataset, DatasetWithShare) and dataset.share:
        return f'({dataset.share.name})'
    return ''


def ask_project(projects: List[Project], input_value: str) -> str:
    project_names = sorted([project.name for project in projects])
    if len(project_names) <= 10:
        return ask(
            'select',
            'What project is this dataset associated with?',
            choices=project_names,
            default=input_value if input_value in project_names else None
        )
    else:
        return ask(
            'autocomplete',
            'What project is this dataset associated with? (use TAB to display options)',
            choices=project_names,
            default=input_value if input_value in project_names else ''
        )


def ask_dataset(datasets: List[Dataset], input_value: str, msg_action: str) -> str:
    if len(datasets) == 0:
        raise InputError("No datasets available")
    sorted_datasets = sorted(datasets, key=lambda d: d.created_at, reverse=True)
    dataset_prompt = {
        'type': 'autocomplete',
        'name': 'dataset',
        'message': f'What dataset would you like to {msg_action}? (Press Tab to see all options)',
        'choices': [f'{dataset.name} - {dataset.id}' for dataset in sorted_datasets],
        'meta_information': {
            f'{dataset.name} - {dataset.id}': f'{format_date(dataset.created_at)} {_format_share(dataset)}'
            for dataset in datasets
        },
        'ignore_case': True
    }
    answers = prompt_wrapper(dataset_prompt)
    choice = answers['dataset']
    # Map the answer to a dataset
    for dataset in datasets:
        if f'{dataset.name} - {dataset.id}' == choice:
            return dataset.id

    # The user has made a selection which does not match
    # any of the options available.
    # This is most likely because there was a typo
    if ask(
        'confirm',
        'The selection does match an option available - try again?'
    ):
        return ask_dataset(datasets, input_value, msg_action)
    raise InputError("Exiting - no dataset selected")



from pathlib import Path

from cirro_api_client.v1.models import Dataset, Project

from cirro.cli.interactive.common_args import ask_project, ask_dataset
from cirro.cli.interactive.utils import prompt_wrapper
from cirro.cli.models import ValidateArguments


def ask_directory(input_value: str) -> str:
    directory_prompt = {
        'type': 'path',
        'name': 'directory',
        'only_directories': True,
        'message': 'What local folder would you like to compare data contents for?',
        'default': input_value or str(Path.cwd())
    }

    answers = prompt_wrapper(directory_prompt)
    return answers['directory']


def gather_validate_arguments(input_params: ValidateArguments, projects: list[Project]):
    input_params['project'] = ask_project(projects, input_params.get('project'))
    return input_params


def gather_validate_arguments_dataset(input_params: ValidateArguments, datasets: list[Dataset]):
    input_params['dataset'] = ask_dataset(datasets, input_params.get('dataset'), msg_action='validate')
    input_params['data_directory'] = ask_directory(input_params.get('data_directory'))
    return input_params



from typing import TypedDict, Optional


class DownloadArguments(TypedDict):
    project: str
    dataset: str
    data_directory: str
    interactive: bool
    file: Optional[list[str]]


class UploadArguments(TypedDict):
    name: str
    description: str
    project: str
    data_type: str
    data_directory: str
    include_hidden: bool
    interactive: bool
    file: Optional[list[str]]


class ValidateArguments(TypedDict):
    dataset: str
    project: str
    data_directory: str
    interactive: bool


class ListArguments(TypedDict):
    project: str
    interactive: bool


class CreatePipelineConfigArguments(TypedDict):
    pipeline_dir: str
    output_dir: str
    entrypoint: Optional[str]
    interactive: bool


class UploadReferenceArguments(TypedDict):
    name: str
    reference_type: str
    project: str
    reference_file: list[str]
    interactive: bool



import importlib

import click
import requests
from cirro_api_client.v1.errors import CirroException

from cirro.cli import run_create_pipeline_config, run_validate_folder
from cirro.cli import run_ingest, run_download, run_configure, run_list_datasets
from cirro.cli.controller import handle_error, run_upload_reference
from cirro.cli.interactive.utils import InputError


def check_required_args(args):
    if args.get('interactive'):
        return
    if any(value is None for value in args.values()):
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
        ctx.exit()


@click.group(help="Cirro CLI - Tool for interacting with datasets")
@click.version_option()
def run():
    pass  # Print out help text, nothing to do


@run.command(help='List datasets', no_args_is_help=True)
@click.option('--project',
              help='Name or ID of the project')
@click.option('-i', '--interactive',
              help='Gather arguments interactively',
              is_flag=True, default=False)
def list_datasets(**kwargs):
    check_required_args(kwargs)
    run_list_datasets(kwargs, interactive=kwargs.get('interactive'))


@run.command(help='Download dataset files', no_args_is_help=True)
@click.option('--project',
              help='Name or ID of the project')
@click.option('--dataset',
              help='ID of the dataset')
@click.option('--file',
              help='Relative path of the file(s) to download (optional, can be used multiple times)',
              default=[],
              multiple=True)
@click.option('--data-directory',
              help='Directory to store the files')
@click.option('-i', '--interactive',
              help='Gather arguments interactively',
              is_flag=True, default=False)
def download(**kwargs):
    check_required_args(kwargs)
    run_download(kwargs, interactive=kwargs.get('interactive'))


@run.command(help='Upload and create a dataset', no_args_is_help=True)
@click.option('--name',
              help='Name of the dataset')
@click.option('--description',
              help='Description of the dataset (optional)',
              default='')
@click.option('--project',
              help='Name or ID of the project')
@click.option('--data-type', '--process',
              help='Name or ID of the data type (--process is deprecated)')
@click.option('--data-directory',
              help='Directory you wish to upload')
@click.option('--file',
              help='Relative path of the file(s) to upload (optional, can be used multiple times)',
              default=[],
              multiple=True)
@click.option('-i', '--interactive',
              help='Gather arguments interactively',
              is_flag=True, default=False)
@click.option('--include-hidden',
              help='Include hidden files in the upload (e.g., files starting with .)',
              is_flag=True, default=False)
def upload(**kwargs):
    check_required_args(kwargs)
    run_ingest(kwargs, interactive=kwargs.get('interactive'))


@run.command(help='Validate a dataset exactly matches a local folder', no_args_is_help=True)
@click.option('--dataset',
              help='Name or ID of the dataset')
@click.option('--project',
              help='Name or ID of the project')
@click.option('--data-directory',
              help='Local directory you wish to validate')
@click.option('-i', '--interactive',
              help='Gather arguments interactively',
              is_flag=True, default=False)
def validate(**kwargs):
    check_required_args(kwargs)
    run_validate_folder(kwargs, interactive=kwargs.get('interactive'))


@run.command(help='Upload a reference to a project', no_args_is_help=True)
@click.option('--name',
              help='Name of the reference')
@click.option('--reference-type',
              help='Type of the reference (e.g., Reference Genome (FASTA))')
@click.option('--project',
              help='Name or ID of the project')
@click.option('--reference-file',
              help='Location of reference file(s) to upload (can be used multiple times)',
              multiple=True)
@click.option('-i', '--interactive',
              help='Gather arguments interactively',
              is_flag=True, default=False)
def upload_reference(**kwargs):
    check_required_args(kwargs)
    run_upload_reference(kwargs, interactive=kwargs.get('interactive'))


@run.command(help='Configure authentication')
def configure():
    run_configure()


@run.command(help='Create pipeline configuration files')
@click.option('-p', '--pipeline-dir',
              metavar='DIRECTORY',
              help='Directory containing the pipeline definition files (e.g., WDL or Nextflow)',
              default='.',
              show_default=True)
@click.option('-e', '--entrypoint',
              help=(
                  'Entrypoint WDL file (optional, if not specified, the first WDL file found will be used).'
                  ' Ignored for Nextflow pipelines.'),
              default='main.wdl')
@click.option('-o', '--output-dir',
              help='Directory to store the generated configuration files',
              default='.cirro',
              show_default=True)
@click.option('-i', '--interactive',
              help='Gather arguments interactively',
              is_flag=True, default=False)
def create_pipeline_config(**kwargs):
    check_required_args(kwargs)
    run_create_pipeline_config(kwargs, interactive=kwargs.get('interactive'))


def _check_version():
    """
    Prompts the user to update their package version if needed
    """
    yellow_color = '\033[93m'
    reset_color = '\033[0m'

    try:
        current_version = importlib.metadata.version('cirro')
        response = requests.get("https://pypi.org/pypi/cirro/json")
        response.raise_for_status()
        latest_version = response.json()["info"]["version"]

        if current_version != latest_version:
            print(f"{yellow_color}Warning:{reset_color} Cirro version {current_version} "
                  f"is out of date. Update to {latest_version} with 'pip install cirro --upgrade'.")

    except Exception:
        return


def main():
    try:
        _check_version()
        run()
    except InputError as e:
        handle_error(e)
    except CirroException as e:
        handle_error(e)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()
